'''Module containing definition of the main Client class'''

# Standard library imports
import uuid

# Imports
import requests
import click

# Package imports
from . import constants
from . import errors
from . import util

class Client(object):
    '''Main Client class for accessing Gencove API'''
    def __init__(self, host, device_uuid=None, refresh_token=None):
        self._set_host(host)
        self._set_device_uuid(device_uuid)
        self.set_refresh_token(refresh_token)

    def _set_host(self, host):
        '''Set host

        Defaults to constants.HOST_DEFAULT
        Even though it can be None, we want the user to explicitly define it as 
        None

        Args:
            host (str): host

        Raises:
            TypeError: if host is not text
        '''
        if host is None:
            click.echo('Using default host: {}'.format(constants.HOST_DEFAULT))
            host = constants.HOST_DEFAULT
        if not util.is_text(host):
            raise TypeError('host must be text')
        self._host = host

    def _set_device_uuid(self, device_uuid=None):
        '''Set device UUID

        Defaults to uuid.getnode()

        Args:
            device_uuid (str): device UUID
        '''
        if device_uuid is None:
            device_uuid = uuid.getnode()

        self._device_uuid = device_uuid

    def set_refresh_token(self, refresh_token):
        '''Set a refresh token for auth

        Args:   
            refresh_token (str): new refresh token

        Raises:
            TypeError: is refresh_token is not text
        '''
        if not (refresh_token is None or util.is_text(refresh_token)):
            raise TypeError('refresh_token must be either None or text')
        self._refresh_token = refresh_token
        self._set_jwt(None)

    def get_refresh_token(self):
        '''Get current refresh token'''
        return self._refresh_token

    def _set_jwt(self, jwt):
        '''Set a JWT for auth

        Args:
            jwt (str): new jwt

        Raises:
            TypeError: if jwt is not text
            ValueError: if jwt is not formatted as Header.Payload.Signature
        '''
        if not (jwt is None or util.is_text(jwt)):
            raise TypeError('jwt must be either None or text')
        if util.is_text(jwt) and len(jwt.split('.')) != 3:
            raise ValueError('jwt must consist of three parts separated by "."')
        self._jwt = jwt  

    def url(self, endpoint, keys={}):
        '''Generate URL for endpoint

        Args:
            endpoint (str): endpoint identifier
            keys (dict): dict defining keys in URL to fill in

        Returns:
            str: complete URL for endpoint

        Raises:
            gencove.errors.URLDataIncompleteError: on unknown endpoint or missing 
                incomplete keys
        '''
        try:
            return (self._host+constants.ENDPOINTS[endpoint]).format(**keys)
        except KeyError as e:
            raise errors.URLDataIncompleteError('Please provide key {} for endpoint \'{}\''.format(e, endpoint))

    def _process_response(self, response):
        '''Common method to process HTTP responses

        It will attempt to decode JSON and check the response code

        Args:
            response (requests.Response): HTTP response to process

        Returns:
            dict: response.text parsed as JSON

        Raises:
            gencove.errors.HTTPError: on HTTP response with an erroneous 
                status_code
            gencove.errors.NoJSONInResponseError: in case JSON cannot be parsed
        '''

        # This will raise gencove.errors.HTTPError if needed
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            raise errors.HTTPError(e.args[0], response=response)

        # This will raise gencove.errors.NoJSONInResponseError if needed
        try:
            response_dict = response.json()
        except ValueError as e:
            raise errors.NoJSONInResponseError(e, response)

        return response_dict

    def register(self, name, email, password=None, return_response=False):
        '''Register new account with Gencove

        In case password is None, user will be prompted for password.

        Args:
            email (str): email to associate with user
            password (str): password to associate with user
            name (str): name to associate with user
            return_response (bool): whether to return backend response parsed as 
                JSON

        Returns:
            None or dict: None or backend response parsed as JSON, depending on
            return_response

        Raises:
            click.Abort: in case user aborts when prompted for input
            gencove.errors.HTTPError: on HTTP response with an erroneous 
                status_code
            gencove.errors.NoJSONInResponseError: in case JSON cannot be parsed
            gencove.errors.URLDataIncompleteError: on unknown endpoint or missing 
                incomplete keys
            TypeError: in case returned refresh token is not text

        '''
        password = util.prompt_stderr(password, confirmation_prompt=True)

        response = requests.post(
            self.url('register'), 
            json={
                'email': email, 
                'password': password, 
                'name': name, 
                'device_uuid': self._device_uuid
            }, 
            headers=constants.HEADER_JSON
        )

        response_dict = self._process_response(response)

        self.set_refresh_token(response_dict['refresh_token'])

        if return_response:
            return response_dict

    def login(self, email, password=None, return_response=False):
        '''Log into Gencove backend

        In case password is None, user will be prompted for password.

        CAUTION: This will log out any sessions (tokens) the user has with 
        current device_uuid

        Keyword arguments:
            email (str): email associated with user
            password (str): password associated with user
            return_response (bool): whether to return backend response parsed as 
                JSON

        Returns:
            None or dict: None or backend response parsed as JSON, depending on
            return_response

        Raises:
            gencove.errors.HTTPError: on HTTP response with an erroneous 
                status_code
            click.Abort: in case user aborts when prompted for input
            gencove.errors.NoJSONInResponseError: in case JSON cannot be parsed
            gencove.errors.URLDataIncompleteError: on unknown endpoint or missing 
                incomplete keys
            TypeError: in case returned refresh token is not text
        '''
        password = util.prompt_stderr(password)

        response = requests.post(
            self.url('login'), 
            json={
                'email': email, 
                'password': password, 
                'device_uuid': self._device_uuid
            }, 
            headers=constants.HEADER_JSON
          )

        response_dict = self._process_response(response)

        self.set_refresh_token(response_dict['refresh_token'])

        if return_response:
            return response_dict

    def logout(self, return_response=False):
        '''Log out of Gencove

        Invalidates refresh token on backend.

        Args:
            return_response (bool): whether to return backend response parsed as 
                JSON

        Returns:
            None or dict: None or backend response parsed as JSON, depending on
            return_response
        
        Raises:
            gencove.errors.HTTPError: on HTTP response with an erroneous 
                status_code
            gencove.errors.NoJSONInResponseError: in case JSON cannot be parsed
            gencove.errors.NotLoggedInError: if not logged in
            gencove.errors.URLDataIncompleteError: on unknown endpoint or missing 
                incomplete keys
            TypeError: in case returned refresh token is not text
        '''
        self.check_login()
        
        response = requests.post(
            self.url('logout'), 
            json={
                'refresh_token': self.get_refresh_token()
            }, 
            headers=constants.HEADER_JSON
        )

        response_dict = self._process_response(response)

        self.set_refresh_token(None)

        if return_response:
            return response_dict

    def email_confirm(self, confirm_id, return_response=False):
        '''Confirm email with confirmation code

        Args:
            confirm_id (str): email confirmation code (usually received via 
                email)
            return_response (bool): whether to return backend response parsed as 
                JSON

        Returns:
            None or dict: None or backend response parsed as JSON, depending on
            return_response
        
        Raises:
            gencove.errors.HTTPError: on HTTP response with an erroneous 
                status_code
            gencove.errors.NoJSONInResponseError: in case JSON cannot be parsed
            gencove.errors.URLDataIncompleteError: on unknown endpoint or missing 
                incomplete keys
        '''

        response_dict = self._process_response(
            # No need for auth
            requests.put(
                self.url('confirm', {'confirm_id': confirm_id})
            )
        )

        if return_response:
            return response_dict

    def password_forgot_init(self, email, return_response=False):
        '''Trigger sending of forgotten password email

        CAUTION: changing password invalidates all sessions (tokens) for current
        user

        Args:
            email (str): email associated with account
            return_response (bool): whether to return backend response parsed as 
                JSON

        Returns:
            None or dict: None or backend response parsed as JSON, depending on
            return_response
        
        Raises:
            gencove.errors.HTTPError: on HTTP response with an erroneous 
                status_code
            gencove.errors.NoJSONInResponseError: in case JSON cannot be parsed
            gencove.errors.URLDataIncompleteError: on unknown endpoint or missing 
                incomplete keys
        '''
        response_dict = self._process_response(
            requests.post(
              self.url('password_forgot_init'),
              json = {
                  'email': email
              }, 
              headers=constants.HEADER_JSON
            )
        )

        if return_response:
            return response_dict

    def check_login(self):
        '''Check if user is logged in

        Raises:
            gencove.errors.NotLoggedInError: if not logged in
        '''
        if self.get_refresh_token() is None:
            raise errors.NotLoggedInError('Not logged in')

    def _get_jwt(self):
        '''Get JWT from Gencove backend'''
        self.check_login()

        response = requests.post(
            self.url('auth'), 
            json={
                'refresh_token': self.get_refresh_token(),
                'empty': 'e'
            }, 
            headers=constants.HEADER_JSON
        )

        response_dict = self._process_response(response)

        self._set_jwt(response_dict['access_token'])

    def _auth_header(self):
        '''Generate Authorization header for HTTP request'''
        if self._jwt is None:
            self._get_jwt()

        return {'Authorization':'JWT {}'.format(self._jwt)}

    def _auth_req(self, method, *args, **kwargs):
        '''Base method for requests to backend endpoints requiring auth'''
        self.check_login()    

        if 'headers' not in kwargs:
            kwargs['headers'] = {}

        kwargs['headers'].update(self._auth_header())
        response = method(*args, **kwargs)

        if response.status_code == 401:
            # trigger fetching of new JWT
            self._set_jwt(None)
            kwargs['headers'].update(self._auth_header())
            response = method(*args, **kwargs)

        return response

    def _get(self, *args, **kwargs):
        '''HTTP GET to backend endpoints requiring auth'''
        return self._auth_req(requests.get, *args, **kwargs)

    def _post(self, *args, **kwargs):
        '''HTTP POST to backend endpoints requiring auth'''
        return self._auth_req(requests.post, *args, **kwargs)

    def _put(self, *args, **kwargs):
        '''HTTP PUT to backend endpoints requiring auth'''
        return self._auth_req(requests.put, *args, **kwargs)

    def _patch(self, *args, **kwargs):
        '''HTTP PATCH to backend endpoints requiring auth'''
        return self._auth_req(requests.patch, *args, **kwargs)

    def _delete(self, *args, **kwargs):
        '''HTTP DELETE to backend endpoints requiring auth'''
        return self._auth_req(requests.delete, *args, **kwargs)

    def password_change(
            self, password, new_password, new_password_confirm, 
            return_response=False):
        '''Change password

        In case passwords are None, user will be prompted for password.

        CAUTION: changing password invalidates all sessions (tokens) for current
        user

        Args:
            password (str): current password
            new_password (str): new password
            new_password_confirm (str): must match new_password
            return_response (bool): whether to return backend response parsed as 
                JSON

        Returns:
            None or dict: None or backend response parsed as JSON, depending on
            return_response
            
        Raises:
            click.Abort: in case user aborts when prompted for input
            gencove.errors.HTTPError: on HTTP response with an erroneous 
                status_code
            gencove.errors.NoJSONInResponseError: in case JSON cannot be parsed
            gencove.errors.NotLoggedInError: if not logged in
            gencove.errors.URLDataIncompleteError: on unknown endpoint or missing 
                incomplete keys
            TypeError: in case new auth JWT or refresh token are not text
            ValueError: in case new auth JWT is not properly formatted
        '''
        password = util.prompt_stderr(password)
        new_password = util.prompt_stderr(new_password, 'New password')
        new_password_confirm = util.prompt_stderr(
            new_password_confirm, 'Confirm new password')

        response_dict = self._process_response(
            self._put(
                self.url('password_change'),
                json = {
                    'password': password,
                    'new_password': new_password,
                    'new_password_confirm': new_password_confirm,
                    'device_uuid': self._device_uuid
                }
            )
          )

        self.set_refresh_token(response_dict['refresh_token'])

        if return_response:
            return response_dict

    def studies_get(self, return_response=False):
        '''Get all studies owned by user

        Returns:
            list or dict: list of studies or backend response parsed as JSON, 
            depending on return_response
            
        Raises:
            gencove.errors.HTTPError: on HTTP response with an erroneous 
                status_code
            gencove.errors.NoJSONInResponseError: in case JSON cannot be parsed
            gencove.errors.NotLoggedInError: if not logged in
            gencove.errors.URLDataIncompleteError: on unknown endpoint or missing 
                incomplete keys
            TypeError: in case new auth JWT is not text
            ValueError: in case new auth JWT is not properly formatted
        '''
        response_dict = self._process_response(
            self._get(
                self.url('studies')
            )
        )

        if return_response:
            return response_dict
        else:
            return response_dict['studies']

    def study_create(
            self, name, href, description, study_enabled=True, 
            discoverable=True, return_response=False):
        '''Create a new study that will be owned by this user

        Args:
            name (str): study name
            href (str): URL of study description
            description (str): study description
            study_enabled (bool): whether the study is enabled upon creation 
                [optional, default: True]
            discoverable (bool): whether the study is discoverable upon 
                creation [optional, default: True]

        Returns:
            dict: study or backend response parsed as JSON, depending on 
                return_response
            
        Raises:
            gencove.errors.HTTPError: on HTTP response with an erroneous 
                status_code
            gencove.errors.NoJSONInResponseError: in case JSON cannot be parsed
            gencove.errors.NotLoggedInError: if not logged in
            gencove.errors.URLDataIncompleteError: on unknown endpoint or missing 
                incomplete keys
            TypeError: in case new auth JWT is not text
            ValueError: in case new auth JWT is not properly formatted
        '''
        response_dict = self._process_response(
            self._post(
                self.url('studies'),
                json = {
                    'name': name,
                    'href': href,
                    'description': description,
                    'study_enabled': study_enabled,
                    'discoverable': discoverable
                }, 
                headers=constants.HEADER_JSON
            )
        )

        if return_response:
            return response_dict
        else:
            return response_dict['study']

    def study_get(self, study_id, return_response=False):
        '''Get study info

        Args:
            study_id (int): study id

        Returns:
            dict: study or backend response parsed as JSON, depending on 
                return_response
            
        Raises:
            gencove.errors.HTTPError: on HTTP response with an erroneous 
                status_code
            gencove.errors.NoJSONInResponseError: in case JSON cannot be parsed
            gencove.errors.NotLoggedInError: if not logged in
            gencove.errors.URLDataIncompleteError: on unknown endpoint or missing 
                incomplete keys
            TypeError: in case new auth JWT is not text
            ValueError: in case new auth JWT is not properly formatted
        '''
        response_dict = self._process_response(
            self._get(
                self.url('study', {'study_id': study_id})
            )
        )

        if return_response:
            return response_dict
        else:
            return response_dict['study']

    def study_update(
            self, study_id, name=None, href=None, description=None, 
            subsidy_enabled=None, study_enabled=None, 
            discoverable=None, return_response=False):
        '''Update study info

        Args:
            study_id (int): study id
            name (str): study name
            href (str): URL of study description
            description (str): study description
            subsidy_enabled (bool): whether subsidies are enabled for this study
            study_enabled (bool): whether the study is enabled
            discoverable (bool): whether the study is discoverable

        Returns:
            dict: study or backend response parsed as JSON, depending on 
                return_response
            
        Raises:
            gencove.errors.HTTPError: on HTTP response with an erroneous 
                status_code
            gencove.errors.NoJSONInResponseError: in case JSON cannot be parsed
            gencove.errors.NotLoggedInError: if not logged in
            gencove.errors.URLDataIncompleteError: on unknown endpoint or missing 
                incomplete keys
            gencove.errors.NoDataError: when all arguments but study_id are None
            TypeError: in case new auth JWT is not text
            ValueError: in case new auth JWT is not properly formatted
        '''    
        update_data = {}
        if name is not None:
            update_data['name'] = name

        if href is not None:
            update_data['href'] = href

        if description is not None:
            update_data['description'] = description

        if subsidy_enabled is not None:
            update_data['subsidy_enabled'] = subsidy_enabled

        if study_enabled is not None:
            update_data['study_enabled'] = study_enabled

        if discoverable is not None:
            update_data['discoverable'] = discoverable

        if len(update_data) == 0:
            raise errors.NoDataError('No data provided. Provide at least one of: name, href, description.')

        response_dict = self._process_response(
            self._put(
                self.url('study', {'study_id': study_id}),
                json = update_data, 
                headers=constants.HEADER_JSON
            )
        )

        if return_response:
            return response_dict
        else:
            return response_dict['study']

    def study_participants_get(self, study_id, return_response=False):
        '''Get a list of study participants

        Args:
            study_id (int): study id

        Returns:
            list or dict: list of participants or backend response parsed as 
            JSON, depending on return_response
            
        Raises:
            gencove.errors.HTTPError: on HTTP response with an erroneous 
                status_code
            gencove.errors.NoJSONInResponseError: in case JSON cannot be parsed
            gencove.errors.NotLoggedInError: if not logged in
            gencove.errors.URLDataIncompleteError: on unknown endpoint or missing 
                incomplete keys
            TypeError: in case new auth JWT is not text
            ValueError: in case new auth JWT is not properly formatted
        '''
        response_dict = self._process_response(
            self._get(
                self.url('study_participants', {'study_id': study_id})
            )
        )

        if return_response:
            return response_dict
        else:
            return response_dict['participants']

    def study_raw_data_get(
            self, study_id, participant_ids=None, return_response=False):
        '''Get presigned links for downloading study data of participants

        Args:
            study_id (int): study id
            participant_ids (list): list of participant ids

        Returns:
            list or dict: list of raw data info or backend response parsed as 
            JSON, depending on return_response
            
        Raises:
            gencove.errors.HTTPError: on HTTP response with an erroneous 
                status_code
            gencove.errors.NoJSONInResponseError: in case JSON cannot be parsed
            gencove.errors.NotLoggedInError: if not logged in
            gencove.errors.URLDataIncompleteError: on unknown endpoint or missing 
                incomplete keys
            TypeError: in case new auth JWT is not text
            ValueError: in case new auth JWT is not properly formatted
        '''
        response_dict = self._process_response(
            self._post(
                self.url('study_raw_data', {'study_id': study_id}),
                json = {
                    'participant_ids': participant_ids
                }, 
                headers=constants.HEADER_JSON
            )
        )

        if return_response:
            return response_dict
        else:
            return response_dict['raw_data']

    def _frontend_version_min_get(self):
        '''Get minimum frontend version

        Currently used for testing

        Returns:
            dict: backend response parsed as JSON
            
        Raises:
            gencove.errors.HTTPError: on HTTP response with an erroneous 
                status_code
            gencove.errors.NoJSONInResponseError: in case JSON cannot be parsed
            gencove.errors.NotLoggedInError: if not logged in
            gencove.errors.URLDataIncompleteError: on unknown endpoint or missing 
                incomplete keys
            TypeError: in case new auth JWT is not text
            ValueError: in case new auth JWT is not properly formatted
        '''
        return self._process_response(
            self._get(
                self.url('frontend_version_min'),
                headers=constants.HEADER_JSON
            )
        )



'''Modules containing project-wide constants'''

HOST_DEFAULT = 'https://rest.gencove.com'
'''Default backend host'''

DEEPLINK_URL_STUDY = 'https://dl.gencove.com/research/{jwt}'
'''URL for deeplinking studies'''

DEEPLINK_URL_APP = 'https://dl.gencove.com/connect/{jwt}'
'''URL for deeplinking apps'''

DEEPLINK_URL_DEFAULT = DEEPLINK_URL_STUDY

_production_env = {
    'host': HOST_DEFAULT
}

ENVS = {
    'production': _production_env,
    'default': _production_env
}
'''Dictionary containing available environments'''

ENDPOINTS = {
    'register': '/register',
    'login': '/login',
    'logout': '/logout',
    'auth': '/auth',
    'password_change': '/password',
    'password_forgot_init': '/password/forgot',
    'studies': '/studies',
    'study': '/studies/{study_id}',
    'study_participants': '/studies/{study_id}/participants',
    'study_raw_data': '/studies/{study_id}/raw-data',
    'confirm': '/confirm/{confirm_id}',
    'frontend_version_min': '/frontend_version_min'
}
'''Dictionary listing REST endpoints available through API'''

HEADER_JSON = {'Content-Type': 'application/json'}
'''HTTP header template for application/json Content-Type'''



# Standard library imports
import json

# Imports
import click

# Package imports
from . import client
from . import errors
from . import constants
from . import util

HOST = 'GENCOVE_HOST'
HOST_HELP = 'backend host, defaults to ${} -> "{}"'.format(HOST, constants.HOST_DEFAULT)
DEVICE_UUID = 'GENCOVE_DEVICE_UUID'
DEVICE_UUID_HELP = 'unique id for client device, defaults to hardware address'
TOKEN = 'GENCOVE_TOKEN'
TOKEN_HELP = 'backend access token, defaults to ${}'.format(TOKEN)

@click.group()
def cli():
    pass

@cli.group()
def auth():
    pass

@cli.group()
def study():
    pass

@study.group()
def participant():
    pass

@auth.command()
@click.argument('name')
@click.argument('email')
@click.option('--password', default=None, help='password associated with user acccount')
@click.option('--host', default=None, help=HOST_HELP, envvar=HOST)
@click.option('--device-uuid', default=None, help=DEVICE_UUID_HELP, envvar=DEVICE_UUID)
@click.option('--print-token/--no-print-token', default=True, help='indicate backend access token should be printed to console')
@click.option('--print-response/--no-print-response', default=False, help='indicate server response should be printed to console')
def register(name, email, password, host, device_uuid, print_token, print_response):
    '''Register new user on gencove backend'''
    c = client.Client(host, device_uuid)

    try:
        response_dict = c.register(name, email, password, return_response=True)
    except errors.HTTPError as e:
        text_out = util.extract_backend_message(e.response)
        click.echo(text_out, err=True)
        exit(1)
    except errors.GencoveClientError as e:
        click.echo(e, err=True)
        exit(1)

    if print_response:
        click.echo(util.get_output(response_dict))
    if print_token:
        click.echo(c.get_refresh_token())

@auth.command()
@click.argument('email')
@click.option('--password', default=None, help='password associated with user acccount')
@click.option('--host', default=None, help=HOST_HELP, envvar=HOST)
@click.option('--device-uuid', default=None, help=DEVICE_UUID_HELP, envvar=DEVICE_UUID)
@click.option('--print-token/--no-print-token', default=True, help='indicate backend access token should be printed to console')
@click.option('--print-response/--no-print-response', default=False, help='indicate server response should be printed to console')
def login(email, password, host, device_uuid, print_token, print_response):
    '''Log into gencove backend

    CAUTION: This will log out any sessions (tokens) the user has with specified device_uuid
    '''
    c = client.Client(host, device_uuid)

    try:
        response_dict = c.login(email, password, return_response=True)
    except errors.HTTPError as e:
        text_out = util.extract_backend_message(e.response)
        click.echo(text_out, err=True)
        exit(1)
    except errors.GencoveClientError as e:
        click.echo(e, err=True)
        exit(1)

    if print_response:
        click.echo(util.get_output(response_dict))
    if print_token:
        click.echo(c.get_refresh_token())

@auth.command()
@click.option('--token', required=True, help=TOKEN_HELP, envvar=TOKEN)
@click.option('--host', default=None, help=HOST_HELP, envvar=HOST)
@click.option('--device-uuid', default=None, help=DEVICE_UUID_HELP, envvar=DEVICE_UUID)
def logout(token, host, device_uuid):
    '''Log out of gencove backend'''
    c = client.Client(host, device_uuid)
    c.set_refresh_token(token)

    try:
        response_dict = c.logout(return_response=True)
    except errors.HTTPError as e:
        text_out = util.extract_backend_message(e.response)
        click.echo(text_out, err=True)
        exit(1)
    except errors.GencoveClientError as e:
        click.echo(e, err=True)
        exit(1)

    click.echo(util.get_output(response_dict))

@auth.command()
@click.argument('confirm-id')
@click.option('--host', default=None, help=HOST_HELP, envvar=HOST)
@click.option('--device-uuid', default=None, help=DEVICE_UUID_HELP, envvar=DEVICE_UUID)
def email_confirm(confirm_id, host, device_uuid):
    '''Confirm email'''
    c = client.Client(host, device_uuid)

    try:
        response_dict = c.email_confirm(confirm_id, return_response=True)
    except errors.HTTPError as e:
        text_out = util.extract_backend_message(e.response)
        click.echo(text_out, err=True)
        exit(1)
    except errors.GencoveClientError as e:
        click.echo(e, err=True)
        exit(1)

    click.echo(util.get_output(response_dict))

@auth.command()
@click.argument('email')
@click.option('--host', default=None, help=HOST_HELP, envvar=HOST)
@click.option('--device-uuid', default=None, help=DEVICE_UUID_HELP, envvar=DEVICE_UUID)
def password_forgot_init(email, host, device_uuid):
    '''Trigger sending of forgotten password email

    CAUTION: changing password invalidates all sessions (tokens) for current user
    '''
    c = client.Client(host, device_uuid)

    try:
        response_dict = c.email_confirm(confirm_id, return_response=True)
    except errors.HTTPError as e:
        text_out = util.extract_backend_message(e.response)
        click.echo(text_out, err=True)
        exit(1)
    except errors.GencoveClientError as e:
        click.echo(e, err=True)
        exit(1)

    click.echo(util.get_output(response_dict))

@auth.command()
@click.option('--token', required=True, help=TOKEN_HELP, envvar=TOKEN)
@click.option('--password', default=None, help='current password associated with user acccount')
@click.option('--new-password', default=None, help='new password to be associated with user acccount')
@click.option('--new-password-confirm', default=None, help='confirmation of new password')
@click.option('--host', default=None, help=HOST_HELP, envvar=HOST)
@click.option('--device-uuid', default=None, help=DEVICE_UUID_HELP, envvar=DEVICE_UUID)
@click.option('--print-token/--no-print-token', default=True, help='indicate new backend access token should be printed to console')
@click.option('--print-response/--no-print-response', default=False, help='indicate server response should be printed to console')
def password_change(token, password, new_password, new_password_confirm, host, device_uuid, print_token, print_response):
    '''Change password

    CAUTION: changing password invalidates all sessions (tokens) for current user
    '''
    c = client.Client(host, device_uuid)
    c.set_refresh_token(token)

    try:
        response_dict = c.password_change(
            password, new_password, new_password_confirm, return_response=True)
    except errors.HTTPError as e:
        text_out = util.extract_backend_message(e.response)
        click.echo(text_out, err=True)
        exit(1)
    except errors.GencoveClientError as e:
        click.echo(e, err=True)
        exit(1)

    if print_response:
        click.echo(util.get_output(response_dict))
    if print_token:
        click.echo(c.get_refresh_token())

@study.command()
@click.option('--token', required=True, help=TOKEN_HELP, envvar=TOKEN)
@click.option('--host', default=None, help=HOST_HELP, envvar=HOST)
@click.option('--device-uuid', default=None, help=DEVICE_UUID_HELP, envvar=DEVICE_UUID)
def get_all(token, host, device_uuid):
    '''List all studies owned by user'''
    c = client.Client(host, device_uuid)
    c.set_refresh_token(token)

    try:
        studies = c.studies_get()
    except errors.HTTPError as e:
        text_out = util.extract_backend_message(e.response)
        click.echo(text_out, err=True)
        exit(1)
    except errors.GencoveClientError as e:
        click.echo(e, err=True)
        exit(1)

    click.echo(util.pretty_json(studies))

@study.command()
@click.argument('study-id')
@click.option('--token', required=True, help=TOKEN_HELP, envvar=TOKEN)
@click.option('--host', default=None, help=HOST_HELP, envvar=HOST)
@click.option('--device-uuid', default=None, help=DEVICE_UUID_HELP, envvar=DEVICE_UUID)
@click.option('--key', default=None, help='output only this key from study info')
def get(study_id, token, host, device_uuid, key):
    '''Get single study owned by user'''
    c = client.Client(host, device_uuid)
    c.set_refresh_token(token)

    try:
        study = c.study_get(study_id)
    except errors.HTTPError as e:
        text_out = util.extract_backend_message(e.response)
        click.echo(text_out, err=True)
        exit(1)
    except errors.GencoveClientError as e:
        click.echo(e, err=True)
        exit(1)

    if key:
        try:
            click.echo(study[key])
        except KeyError as e:
            click.echo('Key "{}" not found'.format(key), err=True)
            exit(1)
    else:
        click.echo(util.pretty_json(study))

@study.command()
@click.argument('name')
@click.argument('href')
@click.argument('description')
@click.option('--enabled/--disabled', default=True, help='whether the study is enabled upon creation')
@click.option('--discoverable/--private', default=True, help='whether the study is discoverable upon creation')
@click.option('--token', required=True, help=TOKEN_HELP, envvar=TOKEN)
@click.option('--host', default=None, help=HOST_HELP, envvar=HOST)
@click.option('--device-uuid', default=None, help=DEVICE_UUID_HELP, envvar=DEVICE_UUID)
@click.option('--key', default=None, help='output only this key from study info')
def create(
        name, href, description, enabled, discoverable, token, host, 
        device_uuid, key):
    '''Create a new study'''
    c = client.Client(host, device_uuid)
    c.set_refresh_token(token)

    try:
        study = c.study_create(
            name=name, href=href, description=description, 
            study_enabled=enabled, discoverable=discoverable)
    except errors.HTTPError as e:
        text_out = util.extract_backend_message(e.response)
        click.echo(text_out, err=True)
        exit(1)
    except errors.GencoveClientError as e:
        click.echo(e, err=True)
        exit(1)

    if key:
        try:
            click.echo(study[key])
        except KeyError as e:
            click.echo('Key "{}" not found'.format(key), err=True)
            exit(1)
    else:
        click.echo(util.pretty_json(study))

@study.command()
@click.argument('study-id')
@click.option('--name', default=None, help='study name')
@click.option('--href', default=None, help='link to study website')
@click.option('--description', default=None, help='study description')
@click.option('--subsidy-enabled/--subsidy-disabled', default=None, help='whether study stubsidies are enabled')
@click.option('--enabled/--disabled', default=None, help='whether the study is enabled')
@click.option('--discoverable/--private', default=None, help='whether the study is discoverable')
@click.option('--token', required=True, help=TOKEN_HELP, envvar=TOKEN)
@click.option('--host', default=None, help=HOST_HELP, envvar=HOST)
@click.option('--device-uuid', default=None, help=DEVICE_UUID_HELP, envvar=DEVICE_UUID)
@click.option('--key', default=None, help='output only this key from study info')
def update(
        study_id, name, href, description, subsidy_enabled, enabled, 
        discoverable, token, host, device_uuid, key):
    '''Update an existing study'''
    c = client.Client(host, device_uuid)
    c.set_refresh_token(token)

    try:
        study = c.study_update(
            study_id=study_id, name=name, href=href, description=description, 
            subsidy_enabled=subsidy_enabled, study_enabled=enabled, 
            discoverable=discoverable)
    except errors.HTTPError as e:
        text_out = util.extract_backend_message(e.response)
        click.echo(text_out, err=True)
        exit(1)
    except errors.GencoveClientError as e:
        click.echo(e, err=True)
        exit(1)

    if key:
        try:
            click.echo(study[key])
        except KeyError as e:
            click.echo('Key "{}" not found'.format(key), err=True)
            exit(1)
    else:
        click.echo(util.pretty_json(study))

@participant.command()
@click.argument('study-id')
@click.option('--token', required=True, help=TOKEN_HELP, envvar=TOKEN)
@click.option('--host', default=None, help=HOST_HELP, envvar=HOST)
@click.option('--device-uuid', default=None, help=DEVICE_UUID_HELP, envvar=DEVICE_UUID)
def get_all(study_id, token, host, device_uuid):
    '''Get all participants of a study'''
    c = client.Client(host, device_uuid)
    c.set_refresh_token(token)

    try:
        participants = c.study_participants_get(study_id)
    except errors.HTTPError as e:
        text_out = util.extract_backend_message(e.response)
        click.echo(text_out, err=True)
        exit(1)
    except errors.GencoveClientError as e:
        click.echo(e, err=True)
        exit(1)

    click.echo(util.pretty_json(participants))

@participant.command()
@click.argument('study-id')
@click.argument('participant-ids', nargs=-1)
@click.option('--token', required=True, help=TOKEN_HELP, envvar=TOKEN)
@click.option('--host', default=None, help=HOST_HELP, envvar=HOST)
@click.option('--device-uuid', default=None, help=DEVICE_UUID_HELP, envvar=DEVICE_UUID)
def get_raw_data(study_id, participant_ids, token, host, device_uuid):
    '''Get all participants of a study'''
    c = client.Client(host, device_uuid)
    c.set_refresh_token(token)

    try:
        raw_data = c.study_raw_data_get(study_id, participant_ids)
    except errors.HTTPError as e:
        text_out = util.extract_backend_message(e.response)
        click.echo(text_out, err=True)
        exit(1)
    except errors.GencoveClientError as e:
        click.echo(e, err=True)
        exit(1)

    click.echo(util.pretty_json(raw_data))

@cli.command()
@click.option('--env', default='default', help='backend environment. Available environments are: {}'.format(util.get_envs_string()))
def host(env):
    if env in constants.ENVS:
        click.echo(constants.ENVS[env]['host'])
    else:
        click.echo('Unknown environment. Available environments are: {}'.format(util.get_envs_string()), err=True)

@cli.command()
@click.argument('study-id', type=int)
@click.argument('external-id')
@click.argument('study-api-key')
@click.option('--subsidy-in-cents', default=None, type=int, help='kit subsidy in cents provided by study')
@click.option('--url-template', default=constants.DEEPLINK_URL_DEFAULT, help='template for generated URL, must contain \'{jwt}\'')
@click.option('--print-url/--no-print-url', default=True, help='print complete deeplink URL or just JWT')
def jwt(study_id, external_id, study_api_key, subsidy_in_cents, url_template, print_url):
    if not print_url:
        url_template = None

    try:
        output = util.jwt_signed(study_id, external_id, study_api_key, subsidy_in_cents, url_template)
    except ValueError as e:
        click.echo(e, err=True)
        exit(1)
    except TypeError as e:
        click.echo(e, err=True)
        exit(1)

    click.echo(output)



from setuptools import setup

import version

setup(
    name='gencove',
    description='Gencove API and CLI tool',
    url='http://docs.gencove.com',
    author='Tomaz Berisa',
    email='tomaz.berisa@gmail.com',
    licence='Apache 2.0',
    version=version.version(),
    packages=['gencove'],
    install_requires=[
        'Click',
        'requests',
        'PyJWT'
    ],
    setup_requires=[
        'pytest-runner'
    ],
    tests_require=[
        'pytest',
        'responses'
    ],
    entry_points='''
        [console_scripts]
        gencove=gencove.cli:cli
    ''',
)